#!/usr/bin/env python

"""
Traverse and annotate simplified AST structures.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from simplified import *
import compiler

class Namespace:
    def __init__(self, local_is_structure=0):
        if local_is_structure:
            self.local = "structure"
        else:
            self.local = "local"
        self.names = {}
        self.not_local = []

    def make_global(self, name):
        if name not in self.not_local:
            self.not_local.append(name)

    def find_for_store(self, name):
        if name not in self.not_local:
            return self.local
        else:
            return "global"

    def find_for_load(self, name):
        if name not in self.not_local and self.names.has_key(name):
            return self.local
        else:
            return "global"

    def store(self, name, types):
        if name not in self.not_local:
            self.names[name] = types
        else:
            raise KeyError, name

    def load(self, name):
        if name in self.not_local or not self.names.has_key(name):
            raise KeyError, name
        else:
            return self.names[name]

    def merge(self, namespace):
        for name, types in namespace.names.items():
            if not self.names.has_key(name):
                self.names[name] = types
            else:
                existing = self.names[name]
                for type in types:
                    if type not in existing:
                        existing.append(type)

class System:
    def __init__(self):
        self.count = 0
    def init(self, node):
        if not hasattr(node, "types"):
            node.types = []
    def annotate(self, node, types):
        for type in types:
            if type not in node.types:
                node.types.append(type)
                self.count += 1

class Annotator(Visitor):
    def __init__(self):
        Visitor.__init__(self)
        self.system = System()
        self.types = None
        self.temp = {}

    def process(self, node, global_namespace=None):
        if hasattr(node, "structure"):
            self.structure = node.structure
            has_structure = 1
        else:
            self.structure = None
            has_structure = 0

        self.namespace = Namespace(self.structure is not None)
        self.global_namespace = global_namespace or self.namespace # NOTE: Improve this.

        if has_structure:
            node.structure.namespace = self.namespace
        node.namespace = self.namespace

        self.visitor = self
        result = self.dispatch(node)

        return result

    def default(self, node):
        for attr in ("args",):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))

        # NOTE: This will eventually use both defaults and supplied arguments.

        for attr in ("params",):
            value = getattr(node, attr, None)
            if value is not None:
                params = []
                for name, default in value:
                    if default is not None:
                        n = self.dispatch(default)
                        self.namespace.store(name, self.types)
                    else:
                        n = None
                        self.namespace.store(name, [])
                    params.append((name, n))
                setattr(node, attr, params)
        for attr in ("expr", "lvalue", "test", "handler", "star", "dstar"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatch(value))
        for attr in ("body", "else_", "finally_", "code"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        return node

    def dispatch(self, node, *args):
        return Visitor.dispatch(self, node, *args)

    def visitGlobal(self, global_):
        for name in global_.names:
            self.namespace.make_global(name)
        return global_

    def visitLoadRef(self, loadref):
        self.types = [loadref.ref]
        return loadref

    def visitLoadName(self, loadname):
        scope = self.namespace.find_for_load(loadname.name)
        print "LoadName", scope
        if scope == "structure":
            return self.dispatch(LoadAttr(expr=LoadRef(ref=self.structure), name=loadname.name))
        elif scope == "global":
            return self.dispatch(LoadGlobal(name=loadname.name))
        else:
            self.types = self.namespace.load(loadname.name)
            return loadname

    def visitStoreName(self, storename):
        scope = self.namespace.find_for_store(storename.name)
        print "StoreName", scope
        if scope == "structure":
            return self.dispatch(StoreAttr(lvalue=LoadRef(ref=self.structure), name=storename.name, expr=storename.expr))
        elif scope == "global":
            return self.dispatch(StoreGlobal(name=storename.name, expr=storename.expr))
        else:
            storename.expr = self.dispatch(storename.expr)
            self.namespace.store(storename.name, self.types)
            return storename

    def visitLoadGlobal(self, loadglobal):
        self.types = self.global_namespace.load(loadglobal.name)
        return loadglobal

    def visitStoreGlobal(self, storeglobal):
        storeglobal.expr = self.dispatch(storeglobal.expr)

        # NOTE: This may always be a merge operation.

        self.global_namespace.store(storeglobal.name, self.types)
        return storeglobal

    def visitLoadTemp(self, loadtemp):
        index = getattr(loadtemp, "index", None)
        self.types = self.temp[index]
        return loadtemp

    def visitStoreTemp(self, storetemp):
        storetemp.expr = self.dispatch(storetemp.expr)
        index = getattr(storetemp, "index", None)
        self.temp[index] = self.types
        return storetemp

    def visitReleaseTemp(self, releasetemp):
        index = getattr(releasetemp, "index", None)
        del self.temp[index]
        return releasetemp

    def visitLoadAttr(self, loadattr):
        loadattr.expr = self.dispatch(loadattr.expr)
        types = []
        for ref in self.types:
            types += ref.namespace.load(loadattr.name)
        self.types = types
        return loadattr

    def visitStoreAttr(self, storeattr):
        storeattr.expr = self.dispatch(storeattr.expr)
        expr = self.types
        storeattr.lvalue = self.dispatch(storeattr.lvalue)
        for ref in self.types:
            ref.namespace.store(storeattr.name, expr)
        return storeattr

# vim: tabstop=4 expandtab shiftwidth=4
